import numpy as np

def mse_cpu(predicted, target): 
	error = np.sum(np.square(predicted - target))/2.0
	return error

def mse_derivate_cpu(predicted, target):
	return (predicted - target)

def softmax_cpu(z, *args):
	z = np.exp(z)
	sumT = z.sum()
	z = z/sumT
	return z

def softmax_derivate_cpu(z, alpha):
	soft = np.exp(z)
	S = soft.sum()
	beta = (alpha*soft).sum()/S
	soft = soft*(alpha - beta)/S
	return soft

def sigmoid2_cpu(z):
	return 2.0*(1.0/(1.0 + np.exp(-z))) - 1.0 # (-1,1)

def sigmoid2_derivate_cpu(z, alpha):
    return alpha*(2.0*np.exp(-z)/((1.0 + np.exp(-z))*(1.0 + np.exp(-z))))

def dotMatrix_cpu(x, w, b):
	return x.dot(w) + b

def dotMatrix_derivate_cpu(w, alpha):
	return alpha.dot(w.transpose())

def transposeDot_cpu(x, derror):
	return x.transpose().dot(derror)

def updateWeights_cpu(w, eta, nabla):
	w = w - (eta * nabla)
	return w


from .activation import activation
from .dactivation import dactivation
from .dlayer import dlayer
from .dloss import dloss
from .dselector import dselector
from .layer import layer
from .loss import loss
from .selector import selector
from .transpose import transpose
from .updateWeights import updateWeight
from .vars import get, reset


import numpy as np
from .lib.GPU import *
from timeit import default_timer as timer
from colorama import Fore, init
from .kernel import *
from .transfer import loadTo
from .brain import Wrapper, Builder

init()
MINIMUMBLOCKSIZE = 28

def ceil(A, B):
    return (A + B - 1) // B

class Neural(object):

    def __init__(self, sizes=None, brain_path=None, eta=0.01, gpu=False):
        if not sizes and not brain_path:
            raise TypeError('Should set `sizes` or `brain_path` params')
        
        self.__eta = np.array([eta])
        self.__num_layers = None
        self.__architecture = None
        self.__weights = None
        self.__biases = None
        self.__gpuMode = gpu
        self.__logs = {}
        self.__mapper = {}

        t = timer()
        self.__setBrain(Builder(brain_path) if brain_path else None, sizes)
        self.__reserveMemo()
        t = timer() - t
    
    def __setBrain(self, brain, sizes):
        if brain:
            self.__num_layers = brain.size()
            self.__architecture = brain.architecture()
            self.__biases = brain.biases()
            self.__weights = brain.weights()
        else:
            self.__num_layers = len(sizes)
            self.__architecture = sizes
            self.__biases = [np.random.uniform(0, -1, x).reshape((1,x)) for x in self.__architecture[1:]]
            self.__weights = [np.random.uniform(-2,2,x*y).reshape((x, y)) for x,y in zip(self.__architecture[:-1], self.__architecture[1:])]

        if not self.__gpuMode:
            return None

        self.__eta = cuda.to_device(self.__eta)
        cuda_biases = [cuda.to_device(bias) for bias in self.__biases]
        self.__biases = cuda_biases

        cuda_weights = [cuda.to_device(weight) for weight in self.__weights]
        self.__weights = cuda_weights
    
    def __getReserve(self, kernel, pos):
        if self.__gpuMode == False:
            return None
        
        return self.__buffer[kernel][pos]
    
    def __reserveArr(self, module_name, shape, streams):
        arr = cuda.device_array(shape, dtype=np.float64)

        stream = cuda.stream()
        self.__buffer[module_name].append(cuda.to_device(arr, stream=stream))
        streams.append(stream)

    def __reserveActivation(self, streams):
        for size in self.__architecture:
            self.__reserveArr('activation', (1, size), streams)
    
    def __reserveLayer(self, streams):
        for size in self.__architecture[1:]:
            self.__reserveArr('layer', (1, size), streams)
    
    def __reserveSelector(self, streams):
        self.__reserveArr('selector', (1, self.__architecture[-1]), streams)
    
    def __reserveDLoss(self, streams):
        self.__reserveArr('d_loss', (1, self.__architecture[-1]), streams)
    
    def __reserveDSelector(self, streams):
        self.__reserveArr('d_selector', (1, self.__architecture[-1]), streams)

    def __reserveDActivation(self, streams):
        LEN = len(self.__architecture)

        for i in range(LEN-1, -1, -1):
            self.__reserveArr('d_activation', (1, self.__architecture[i]), streams)
    
    def __reserveTranspose(self, streams):
        LEN = len(self.__architecture)

        for i in range(LEN-2, -1, -1):
            self.__reserveArr('transpose', (self.__architecture[i], self.__architecture[i+1]), streams)
    
    def __reserveDLayer(self, streams):
        LEN = len(self.__architecture)

        for i in range(LEN-2, -1, -1):
            self.__reserveArr('d_layer', (1, self.__architecture[i]), streams)

    def __reserveExtra(self, streams):
        self.__reserveArr('extra', (1, ), streams)

    def __reserveMemo(self):
        if self.__gpuMode == False:
            return None

        self.__buffer = {
            'activation': [],
            'layer': [],
            'selector': [],
            'd_loss': [],
            'd_selector': [],
            'd_activation': [],
            'transpose': [],
            'd_layer': [],
            'extra': [],
        }

        streams = []
        self.__reserveActivation(streams)
        self.__reserveLayer(streams)
        self.__reserveSelector(streams)
        self.__reserveDLoss(streams)
        self.__reserveDSelector(streams)
        self.__reserveDActivation(streams)
        self.__reserveTranspose(streams)
        self.__reserveDLayer(streams)
        self.__reserveExtra(streams)

        for stream in streams:
            stream.synchronize()

    def logs(self):
        return self.__logs

    def __logger(self, method, delta, dbg=False):
        delta = round(delta * 1000, 3)

        mode = '[GPU] ' if self.__gpuMode else '[CPU] '
        name = mode + method

        if name not in self.__logs.keys():
            self.__logs[name] = []
        self.__logs[name].append(delta)

        color = Fore.GREEN
        if delta >= 1.0:
            color = Fore.YELLOW
        if delta >= 5.0:
            color = Fore.RED
        
        if dbg:
            print(color + "{}: {}ms".format(method, delta))

    def __swapper(self, *args, buffer, GPURunner, CPURunner):
        arr = None
        t = timer()

        name = GPURunner.__name__
        dbg = True if GPURunner.__name__ == 'transpose' else False

        if self.__gpuMode == False:
            arr = CPURunner(*args)
        else:
            mapped_args = []
            for arg in args:
                if cuda.is_cuda_array(arg):
                    mapped_args.append(arg)
                else:
                    arg_dvc = cuda.to_device(arg)
                    mapped_args.append(arg_dvc)
                    self.__mapper[id(arg)] = arg_dvc
                    print("register")
                    assert False

            arr = GPURunner(*mapped_args, buffer)

        t = timer() - t
        self.__logger(name, t, dbg=False)

        return arr

    def __loss(self, predicted, target, buffer=None):
        return self.__swapper(predicted, target, buffer=buffer, GPURunner=loss, CPURunner=mse_cpu)

    def __d_loss(self, predicted, target, buffer=None):
        return self.__swapper(predicted, target, buffer=buffer, GPURunner=dloss, CPURunner=mse_derivate_cpu)

    def __selector(self, z, buffer=None):
        return self.__swapper(z, self.__getReserve('extra', 0), buffer=buffer, GPURunner=selector, CPURunner=softmax_cpu)

    def __d_selector(self, z, alpha, buffer=None):
        return self.__swapper(z, alpha, buffer=buffer, GPURunner=dselector, CPURunner=softmax_derivate_cpu)

    def __activation(self, z, buffer=None):
       return self.__swapper(z, buffer=buffer, GPURunner=activation, CPURunner=sigmoid2_cpu)

    def __d_activation(self, z, alpha, buffer=None):
        return self.__swapper(z, alpha, buffer=buffer, GPURunner=dactivation, CPURunner=sigmoid2_derivate_cpu)
    
    def __layer(self, x, w, b, buffer=None):
        return self.__swapper(x, w, b, buffer=buffer, GPURunner=layer, CPURunner=dotMatrix_cpu)

    def __d_layer(self, _, w, alpha, buffer=None):
        return self.__swapper(w, alpha, buffer=buffer, GPURunner=dlayer, CPURunner=dotMatrix_derivate_cpu)
    
    def __updateWeight(self, weights, eta, nabla_w, buffer=None):
        return self.__swapper(weights, eta, nabla_w, buffer=buffer, GPURunner=updateWeight, CPURunner=updateWeights_cpu)
    
    def __transpose(self, z, derror, buffer=None):
        return self.__swapper(z, derror, buffer=buffer, GPURunner=transpose, CPURunner=transposeDot_cpu)

    def __feedForward(self, x):
        t = timer()

        activation_pointer = 0
        layer_pointer = 0

        for w, b in zip(self.__weights,self.__biases):
            arr = self.__activation(x, buffer=self.__getReserve('activation', activation_pointer))

            x = self.__layer(arr, w, b, buffer=self.__getReserve('layer', layer_pointer))

            activation_pointer += 1
            layer_pointer += 1

        y = self.__selector(x, buffer=self.__getReserve('selector', 0))

        t = timer() - t
        self.__logger("feedForward", t)
        return y

    def __backPropagation(self, x, target):
        t = timer()

        # feedForward
        z = [x] # save all Zs
        activations = [] # save all activations

        activation_pointer = 1
        layer_pointer = 0

        for w, b in zip(self.__weights, self.__biases):
            x = self.__layer(x, w, b, buffer=self.__getReserve('layer', layer_pointer))
            layer_pointer += 1

            activations.append(x)
            x = self.__activation(x, buffer=self.__getReserve('activation', activation_pointer))
            activation_pointer += 1

            z.append(x)

        y = self.__selector(x, buffer=self.__getReserve('selector', 0))

        derror = self.__d_loss(y, target, buffer=self.__getReserve('d_loss', 0))

        derror = self.__d_selector(z[self.__num_layers - 1], derror, buffer=self.__getReserve('d_selector', 0))

        d_activation_pointer = 0
        d_layer_pointer = 0
        transpose_pointer = 0

        for l in range(1, self.__num_layers):
            w = self.__weights[-l]
            b = self.__biases[-l]

            derror = self.__d_activation(activations[-l], derror, buffer=self.__getReserve('d_activation', d_activation_pointer))

            d_activation_pointer += 1
            
            nabla_w = self.__transpose(z[-l-1], derror, buffer=self.__getReserve('transpose', transpose_pointer))
            transpose_pointer += 1

            nabla_b = derror # error for each bias

            derror =  self.__d_layer(z[-l-1], w, derror, buffer=self.__getReserve('d_layer', d_layer_pointer))

            d_layer_pointer += 1

            self.__weights[-l] = self.__updateWeight(self.__weights[-l], self.__eta, nabla_w)

            self.__biases[-l] = self.__updateWeight(self.__biases[-l], self.__eta, nabla_b)

        t = timer() - t
        prefix = '[GPU] ' if self.__gpuMode else '[CPU] '
        self.__logger(prefix + "backpropagation", t, dbg=False)
    
    def __buildMsg(self, x):
        if self.__gpuMode:
            return cuda.to_device(x)
        return x

    def send(self, l):
        t = timer()

        x =  self.__buildMsg(np.array([l]))
        arr = self.__feedForward(x)

        hst, = loadTo(arr, mode='CPU')
        y = hst[0]

        t = timer() - t
        self.__logger("send", t)

        return y

    def learn(self, x, y):
        t = timer()

        x = self.__activation(self.__buildMsg(np.array([x])), buffer=self.__getReserve('activation', 0))
        y = self.__buildMsg(np.array([y]))

        self.__backPropagation(x, y)

        t = timer() - t
        self.__logger("learn", t)

    def cost(self, x, y):
        t = timer()

        np_x = self.__buildMsg(np.array([x]))
        np_y = self.__buildMsg(np.array([y]))
        np_x = self.__activation(np_x, buffer=self.__getReserve('activation', 0))

        ret = self.__loss(self.__feedForward(np_x),np_y)

        t = timer() - t
        self.__logger("cost", t)
        return ret
    
    def export(self, filename, path):
        layers = []

        for w, b in zip(self.__weights, self.__biases):
            w_host, b_host = loadTo(w, b, mode='CPU')
            layers.append((w_host, b_host))
        
        wrapper = Wrapper(layers)

        with open(path + filename + '.brain', 'wb') as file:
            file.write(wrapper.data())



from numba import cuda
from ..lib.GPU import *
from timeit import default_timer as timer
from .vars import add
from ..transfer.loader import loadTo

def dlayer(w, alpha, buffer):
    LEN = w.shape[0]
    LEN2 = w.shape[1]

    t = timer()
    w_dvc, alpha_dvc = loadTo(w, alpha, mode='GPU')
    t = timer() - t
    add(t)

    memset[kernelConfig2D(buffer.shape[0], buffer.shape[1])](buffer)
    cuda.synchronize()
    dotMatrix_derivate[kernelConfig3D(1, LEN, LEN2)](buffer, w_dvc, alpha_dvc)
    cuda.synchronize()

    return buffer


from numba import cuda
from ..lib.GPU import *
from timeit import default_timer as timer
from .vars import add
from ..transfer.loader import loadTo

def selector(z, res, buffer):
    LEN = z.shape[1]

    t = timer()
    z_dvc, res_dvc = loadTo(z, res, mode='GPU')
    memset[kernelConfig1D(1)](res_dvc)
    cuda.synchronize()
    t = timer() - t
    add(t)

    softmax_p1[kernelConfig1D(LEN)](buffer, z_dvc, res_dvc)
    cuda.synchronize()

    softmax_p2[kernelConfig1D(LEN)](buffer, res_dvc)
    cuda.synchronize()

    return buffer


from numba import cuda
from ..lib.GPU import *
from timeit import default_timer as timer
from .vars import add
from ..transfer.loader import loadTo

def updateWeight(weights, eta, nabla_w, _):
    LEN1, LEN2 = weights.shape

    t = timer()
    eta_dvc, nabla_w_dvc, w_dvc = loadTo(eta, nabla_w, weights, mode='GPU')
    t = timer() - t
    add(t)

    updateWeights[kernelConfig2D(LEN1, LEN2)](w_dvc, eta_dvc, nabla_w_dvc)
    cuda.synchronize()

    return w_dvc


MINIMUMBLOCKSIZE = 28
EPS = 1e-10
THREADSPERBLOCK = 1024

def ceil(A, B):
	return (A + B - 1) // B

def kernelConfig1D(size_x):
	threads_x = THREADSPERBLOCK
	blockspergrid_x = max(ceil(size_x, threads_x), MINIMUMBLOCKSIZE)

	return (blockspergrid_x, threads_x)

def kernelConfig2D(size_x, size_y, shape=None):
	threads = THREADSPERBLOCK

	sz = [size_x, size_y]
	t = [1, 1]

	t[0] = 32
	t[1] = 32

	if shape:
		t[0] = shape[0]
		t[1] = shape[1]
	
	blockspergrid_x = ceil(size_x, t[0])
	blockspergrid_y = ceil(size_y, t[1])

	blocks = blockspergrid_x * blockspergrid_y
	if blocks < MINIMUMBLOCKSIZE:
		add = MINIMUMBLOCKSIZE - blocks
		blocks //= blockspergrid_y
		blockspergrid_y += ceil(add, blocks)
	
	blockspergrid = (blockspergrid_x, blockspergrid_y)
	threadsperblock = (t[0], t[1])

	return (blockspergrid, threadsperblock)

def kernelConfig3D(size_x, size_y, size_z):
	threads = THREADSPERBLOCK

	sz = [size_x, size_y, size_z]
	t = [1, 1, 1]

	upd = True
	while threads > 1 and upd:
		upd = False
		for i in range(3):
			if t[i] >= sz[i] or threads == 1:
				continue
			threads //= 2
			t[i] *= 2
			upd = True
	
	blockspergrid_x = ceil(size_x, t[0])
	blockspergrid_y = ceil(size_y, t[1])
	blockspergrid_z = ceil(size_z, t[2])

	blocks = blockspergrid_x * blockspergrid_y * blockspergrid_z
	if blocks < MINIMUMBLOCKSIZE:
		add = MINIMUMBLOCKSIZE - blocks
		blocks //= blockspergrid_z
		blockspergrid_z += ceil(add, blocks)
	
	blockspergrid = (blockspergrid_x, blockspergrid_y, blockspergrid_z)
	threadsperblock = (t[0], t[1], t[2])

	return (blockspergrid, threadsperblock)



from numba import cuda
from ..lib.GPU import *
from timeit import default_timer as timer
from .vars import add
from ..transfer.loader import loadTo

def transpose(z, derror, buffer):
    LEN1 = z.shape[1]
    LEN2 = derror.shape[1]

    t = timer()
    x_dvc, derror_dvc = loadTo(z, derror, mode='GPU')
    t = timer() - t
    add(t)

    transposeDot[kernelConfig2D(LEN1, LEN2)](buffer, x_dvc, derror_dvc)
    cuda.synchronize()

    return buffer


import numpy as np
from . import decoder, encoder

MINIMUM_SIZE = 12
SIGNATURE_SIZE = 7
INT_SIZE = 4
LAYER_HEADER_SIZE = 9
FLOAT_SIZE = 8

class Builder(object): 
    def __init__(self, path):
        self.__size = None
        self.__architecture = []
        self.__biases = []
        self.__weights = []
        self.__error = {
            0: 'type `brain` has an invalid size',
            1: 'type `brain` has not valid signature',
            2: 'data is corrupted. The general security code not match',
            3: 'layer header has no sufficient size',
            4: 'layer target has no match',
            5: 'weight has no sufficient size',
            6: 'weitght target has no match',
            7: 'biase has no sufficient size',
            8: 'biase target has no match',
            9: 'layers are not connect, shapes should be (x, y), (y, z)',
        }
        
        data = self.__read(path)
        self.__build(data)
    
    def __read(self, path):
        data = b''
        with open(path, 'rb') as file:
            data = file.read()
        
        return data
    
    def __build(self, data):
        if len(data) < MINIMUM_SIZE:
            self.__raise(0)
        
        magic_number = decoder.toInt(encoder.securityCode(data[0:-1]))
        if magic_number != data[-1]:
            self.__raise(2)
        
        prefix = 0
        signature = decoder.toString(data[prefix:prefix+SIGNATURE_SIZE])
        if signature != 'kinhosz':
            self.__raise(1)
        
        prefix += SIGNATURE_SIZE
        
        self.__size = decoder.toInt(data[prefix:prefix + INT_SIZE])
        prefix += INT_SIZE
        
        for _ in range(1, self.__size):
            prefix = self.__buildLayer(data, prefix)
        
        for i in range(1, self.__size - 1):
            if self.__weights[i].shape[0] != self.__weights[i-1].shape[1]:
                self.__raise(9)
        
        for i in range(self.__size - 1):
            self.__architecture.append(self.__weights[i].shape[0])
        self.__architecture.append(self.__weights[-1].shape[1])
    
    def __buildLayer(self, data, offset):
        if len(data) < offset + LAYER_HEADER_SIZE:
            self.__raise(3)
        
        target = chr(data[offset])
        if target != 'L':
            self.__raise(4)
        
        offset += 1
        
        dim1 = decoder.toInt(data[offset:offset+INT_SIZE])
        offset += INT_SIZE
        dim2 = decoder.toInt(data[offset:offset+INT_SIZE])
        offset +=  INT_SIZE
        
        offset = self.__buildWeight(data, dim1, dim2, offset)
        offset = self.__buildBiase(data, dim2, offset)
        
        return offset
    
    def __buildWeight(self, data, dim1, dim2, offset):
        REQUIRED_LEN = offset + FLOAT_SIZE * (dim1 * dim2) + 1
        
        if len(data) < REQUIRED_LEN:
            self.__raise(5)
        
        target = chr(data[offset])
        if target != 'W':
            self.__raise(6)
        
        offset += 1
        
        weight = np.empty((dim1, dim2))
        
        for i in range(dim1):
            for j in range(dim2):
                weight[i, j] = decoder.toFloat(data[offset:offset+FLOAT_SIZE])
                offset += FLOAT_SIZE
        
        self.__weights.append(weight)
        
        return offset
    
    def __buildBiase(self, data, dim, offset):
        REQUIRED_LEN = offset + FLOAT_SIZE * dim + 1
        
        if len(data) < REQUIRED_LEN:
            self.__raise(7)
        
        target = chr(data[offset])
        if target != 'B':
            self.__raise(8)
        
        offset += 1
        
        biase = np.empty((1, dim))
        
        for i in range(dim):
            biase[0, i] = decoder.toFloat(data[offset:offset+FLOAT_SIZE])
            offset += FLOAT_SIZE
        
        self.__biases.append(biase)
        
        return offset
    
    def __raise(self, flag):
        raise TypeError(self.__error[flag])
    
    def size(self):
        return self.__size
    
    def architecture(self):
        return self.__architecture
    
    def biases(self):
        return self.__biases
    
    def weights(self):
        return self.__weights


import math
from numba import cuda, float64, float32, int64

DIM1 = (1024, )
DIM2 = (4, 256)

@cuda.jit
def copy(arr, A):
	x = cuda.grid(1)

	if x < arr.shape[1]:
		arr[0, x] = A[0, x]

@cuda.jit
def memset(arr):
	x = cuda.grid(1)

	if x < arr.shape[0]:
		arr[x] = float64(0.)

@cuda.jit
def memset2(arr):
	x, y = cuda.grid(2)

	if x < arr.shape[0] and y < arr.shape[1]:
		arr[x, y] = float64(0.)

@cuda.jit
def mse(result, predicted, target):
    pos = cuda.grid(1)

    if pos < predicted.shape[0]:
        diff = predicted[pos] - target[pos]
        diff = (diff * diff) / 2.0
        cuda.atomic.add(result, 0, diff)

@cuda.jit
def mse_derivate(result, predicted, target):
    pos = cuda.grid(1)

    if pos < predicted.shape[1]:
        result[0, pos] = predicted[0, pos] - target[0, pos]

@cuda.jit
def softmax_p1(arr, z, res):
	local_sum = cuda.shared.array(shape=DIM1, dtype=float64)

	x = cuda.grid(1)
	tx = cuda.threadIdx.x
	local_sum[tx] = float64(.0)

	if x >= arr.shape[1]:
		return

	val = math.exp(z[0, x])
	arr[0, x] = val
	local_sum[tx] = val
	cuda.syncthreads()

	if tx != 0:
		return
	
	for i in range(1, cuda.blockDim.x):
		local_sum[0] += local_sum[i]

	cuda.atomic.add(res, 0, local_sum[0])

@cuda.jit
def softmax_p2(arr, sumT):
	x = cuda.grid(1)

	if x < arr.shape[1]:
		arr[0, x] = arr[0, x] / sumT[0]

@cuda.jit
def softmax_sum_derivate(arr, z, alpha, simple_sum, sum_times_alpha):
	x = cuda.grid(1)

	if x < arr.shape[1]:
		value = z[0, x]
		value = math.exp(value)
		arr[0, x] = value
		cuda.atomic.add(simple_sum, 0, value)
		cuda.atomic.add(sum_times_alpha, 0, value * alpha[0, x])

@cuda.jit
def softmax_derivate(arr, alpha, simple_sum, sum_times_alpha):
	x = cuda.grid(1)
 
	if x < arr.shape[1]:
		value = arr[0, x]
		arr[0, x] = (value * (alpha[0, x] - (sum_times_alpha[0] / simple_sum[0]))) / simple_sum[0]

@cuda.jit
def sigmoid2(arr, A):
	x = cuda.grid(1)

	if x < arr.shape[1]:
		arr[0, x] = 2.0 * (1.0 / (1.0 + math.exp(-A[0, x]))) - 1.0

@cuda.jit
def sigmoid2_derivate(arr, A, alpha):
	x = cuda.grid(1)

	if x < A.shape[1]:
		value = A[0, x]
		arr[0, x] = alpha[0, x] * (2.0 * math.exp(-value) / ( (1.0 + math.exp(-value)) * (1.0 + math.exp(-value))))

@cuda.jit
def sum(arr, C):
	x, y = cuda.grid(2)

	if x >= arr.shape[0] or y >= arr.shape[1]:
		return

	arr[x, y] += C[x, y]

@cuda.jit
def dotMatrix(arr, A, B):
	x, y = cuda.grid(2)
	
	temp = cuda.shared.array(shape=DIM2, dtype=float64)
	sA = cuda.shared.array(shape=(1, DIM2[0]), dtype=float64)

	tx = cuda.threadIdx.x
	ty = cuda.threadIdx.y
	temp[tx, ty] = float64(.0)

	if x >= B.shape[0] or y >= B.shape[1]:
		return
	
	if ty == 0:
		sA[0, tx] = A[0, x]
	cuda.syncthreads()

	temp[tx, ty] = sA[0, tx] * B[x, y]
	cuda.syncthreads()

	if tx != 0:
		return
	
	for i in range(1, cuda.blockDim.x):
		temp[0, ty] += temp[i, ty]
	cuda.atomic.add(arr, (0, y), temp[0, ty])

@cuda.jit
def dotMatrix_derivate(arr, w, alpha):
	x, y, z = cuda.grid(3)

	if x >= arr.shape[0] or y >= arr.shape[1] or z >= alpha.shape[1]:
		return

	cuda.atomic.add(arr, (x, y), alpha[x, z] * w[y, z])

@cuda.jit
def transposeDot(arr, A, B):
	x, y = cuda.grid(2)

	if x < arr.shape[0] and y < arr.shape[1]:
		arr[x, y] = A[0, x] * B[0, y]

@cuda.jit
def updateWeights(w, eta, nabla):
	x, y = cuda.grid(2)

	if x < w.shape[0] and y < w.shape[1]:
		w[x, y] -= eta[0] * nabla[x, y]


from numba import cuda
from timeit import default_timer as timer

def loadTo(*args, mode):
    ret_args = []
    for arg in args:
        if cuda.is_cuda_array(arg):
            if mode == 'GPU':
                ret_args.append(arg)
            else:
                ret_args.append(arg.copy_to_host())
        else:
            if mode == 'GPU':
                ret_args.append(cuda.to_device(arg))
            else:
                ret_args.append(arg)

    return ret_args





from . import encoder

class Wrapper(object):
    def __init__(self, layers):
        self.__data = b''
        self.__error = {
            0: 'weight has shape different of (x, y)',
            1: 'biase has shape different of (1, x)',
            2: 'weight has second dimension different of biase. Should be: (x, y), (1, y)',
            3: 'values in weight are not float',
            4: 'values in biase are not float',
            5: 'layers dimensions are not connected. Should be: (x, y), (y, z)',
        }
        
        self.__checkType(layers)
        self.__buildData(layers)
    
    def __raise(self, flag):
        raise TypeError(self.__error[flag])
    
    def __checkType(self, layers):
        for w, b in layers:
            if len(w.shape) != 2:
                self.__raise(0)
            if len(b.shape) != 2:
                self.__raise(1)
            
            if w.shape[1] != b.shape[1]:
                self.__raise(2)
            
            for w_x in w:
                for w_x_y in w_x:
                    if not isinstance(w_x_y, float):
                        self.__raise(3)
            
            for b_x in b[0]:
                if not isinstance(b_x, float):
                    self.__raise(4)
        
        for i in range(1, len(layers)):
            if layers[i][0].shape[0] != layers[i-1][0].shape[1]:
                self.__raise(5)
    
    def __buildData(self, layers):
        header = b''
        header += encoder.fromString('kinhosz')
        header += encoder.fromInt(len(layers) + 1)
        
        body = b''
        
        for w, b in layers:
            layer = b''
            layer += encoder.fromString('L')
            layer += encoder.fromInt(w.shape[0])
            layer += encoder.fromInt(w.shape[1])

            layer += encoder.fromString('W')
            for w_x in w:
                for w_x_y in w_x:
                    layer += encoder.fromFloat(w_x_y)
            
            layer += encoder.fromString('B')
            for b_x in b[0]:
                layer += encoder.fromFloat(b_x)
            
            body += layer
        
        data = header + body
        
        security_code = encoder.securityCode(data)
        
        data += security_code
        self.__data = data
    
    def data(self):
        return self.__data




from numba import cuda
from ..lib.GPU import *
from timeit import default_timer as timer
from .vars import add
from ..transfer.loader import loadTo
from colorama import Fore

def layer(x, w, b, buffer):
    LEN = w.shape[0]
    LEN2 = w.shape[1]

    t = timer()
    x_dvc, w_dvc, b_dvc = loadTo(x, w, b, mode='GPU')
    t = timer() - t
    add(t)

    copy[kernelConfig1D(buffer.shape[1])](buffer, b_dvc)
    cuda.synchronize()
    dotMatrix[kernelConfig2D(LEN, LEN2, shape=(4, 256))](buffer, x_dvc, w_dvc)
    cuda.synchronize()

    return buffer



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import re
from setuptools import find_packages, setup

package = 'Kinho'

with open(os.path.join(package, '__init__.py'), 'rb') as f:
    init_py = f.read().decode('utf-8')

version = re.search(
    '^__version__ = [\'\"]([^\'\"]+)[\'\"]', init_py, re.MULTILINE
).group(1)
author = re.search(
    '^__author__ = [\'\"]([^\'\"]+)[\'\"]', init_py, re.MULTILINE
).group(1)
email = re.search(
    '^__email__ = [\'\"]([^\'\"]+)[\'\"]', init_py, re.MULTILINE
).group(1)

setup(
    name='Kinho',
    packages=find_packages(),
    version=version,
    description='A library to classify images with deep learning.',
    long_description='The library features the "Neural" model, which is a Convolutional Neural ' \
        + 'Network (CNN) for image classification. It supports both CPU and GPU, providing excellent performance ' \
        + 'for large networks. You can also export and import Deep models and continue training on other machines. ' \
        + 'The exported file is generated in the ".brain" format, which is a proprietary data type of this project. '\
        + 'For more information, please visit our repository.',
    author=author,
    author_email=email,
    url='https://github.com/kinhosz/Neural',
    install_requires=[],
    license='MIT',
    keywords=['dev', 'web'],
    classifiers=[
       'Development Status :: 5 - Production/Stable',
       'Environment :: GPU',
       'Environment :: GPU :: NVIDIA CUDA',
       'Environment :: GPU :: NVIDIA CUDA :: 12 :: 12.0',
       'License :: OSI Approved :: MIT License',
       'Natural Language :: Portuguese (Brazilian)',
       'Natural Language :: English',
       'Programming Language :: Python',
       'Programming Language :: Python :: 3',
       'Topic :: Scientific/Engineering :: Artificial Intelligence',
       'Topic :: Scientific/Engineering :: Image Processing',
       'Topic :: Scientific/Engineering :: Image Recognition',
    ],
)


from numba import cuda
from ..lib.GPU import *
from timeit import default_timer as timer
from .vars import add
from ..transfer.loader import loadTo

def dselector(z, alpha, buffer):
    LEN = z.shape[1]

    t = timer()
    z_dvc, alpha_dvc, ss_dvc, st_dvc = loadTo(z, alpha, np.zeros(1), np.zeros(1), mode='GPU')
    t = timer() - t
    add(t)

    softmax_sum_derivate[kernelConfig1D(LEN)](buffer, z_dvc, alpha_dvc, ss_dvc, st_dvc)
    cuda.synchronize()
    softmax_derivate[kernelConfig1D(LEN)](buffer, alpha_dvc, ss_dvc, st_dvc)
    cuda.synchronize()

    return buffer



from numba import cuda
from ..lib.GPU import *
from timeit import default_timer as timer
from .vars import add
from ..transfer.loader import loadTo

def dactivation(z, alpha, buffer):
    LEN = z.shape[1]

    t = timer()
    z_dvc, alpha_dvc = loadTo(z, alpha, mode='GPU')
    t = timer() - t
    add(t)

    sigmoid2_derivate[kernelConfig1D(LEN)](buffer, z_dvc, alpha_dvc)
    cuda.synchronize()

    return buffer


from numba import cuda
from ..lib.GPU import *
from timeit import default_timer as timer
from .vars import add
from ..transfer.loader import loadTo

def loss(predicted, target, buffer):
    t = timer()
    predicted_dvc, target_dvc = loadTo(predicted, target, mode='GPU')
    t = timer() - t
    add(t)

    mse[kernelConfig1D(predicted_dvc.shape[0])](buffer, predicted_dvc, target_dvc)
    cuda.synchronize()

    return buffer


from numba import cuda
from ..lib.GPU import *
from timeit import default_timer as timer
from .vars import add
from ..transfer.loader import loadTo

def dloss(predicted, target, buffer):
    LEN = predicted.shape[1]

    t = timer()
    predicted_dvc, target_dvc = loadTo(predicted, target, mode='GPU')
    t = timer() - t
    add(t)

    mse_derivate[kernelConfig1D(LEN)](buffer, predicted_dvc, target_dvc)
    cuda.synchronize()

    return buffer


acm = 0
calls = 0
avg = 0.0

def add(t):
    global acm

    acm += t

def get():
    global acm
    return acm

def reset():
    global acm
    global calls
    global avg

    avg = 0.0
    acm = 0
    calls = 0